import time
import logging
import numpy as np
from scipy.sparse.linalg import svds
from sklearn.utils.extmath import randomized_svd


def pcp(M, delta=1e-6, mu=None, maxiter=500, verbose=False, missing_data=True,
        svd_method="approximate", **svd_args):
    # Check the SVD method.
    allowed_methods = ["exact", "sparse", "randomized"]
    if svd_method not in allowed_methods:
        raise ValueError("'svd_method' must be one of: {0}"
                         .format(allowed_methods))

    # Check for missing data.
    shape = M.shape
    if missing_data:
        missing = ~(np.isfinite(M))
        if np.any(missing):
            M = np.array(M)
            M[missing] = 0.0
    else:
        missing = np.zeros_like(M, dtype=bool)
        if not np.all(np.isfinite(M)):
            logging.warn("The matrix has non-finite entries. "
                         "SVD will probably fail.")

    # Initialize the tuning parameters.
    lam = 1.0 / np.sqrt(np.max(shape))
    if mu is None:
        mu = 0.25 * np.prod(shape) / np.sum(np.abs(M))
        if verbose:
            print("mu = {0}".format(mu))

    # Convergence criterion.
    norm = np.sum(M ** 2)

    # Iterate.
    i = 0
    rank = np.min(shape)
    S = np.zeros(shape)
    Y = np.zeros(shape)
    while i < max(maxiter, 1):
        # SVD step.
        strt = time.time()
        u, s, v = _svd(svd_method, M - S + Y / mu, rank+1, 1./mu, **svd_args)
        svd_time = time.time() - strt

        s = shrink(s, 1./mu)
        rank = np.sum(s > 0.0)
        u, s, v = u[:, :rank], s[:rank], v[:rank, :]
        L = np.dot(u, np.dot(np.diag(s), v))

        # Shrinkage step.
        S = shrink(M - L + Y / mu, lam / mu)

        # Lagrange step.
        step = M - L - S
        step[missing] = 0.0
        Y += mu * step

        # Check for convergence.
        err = np.sqrt(np.sum(step ** 2) / norm)
        if verbose:
            print(("Iteration {0}: error={1:.3e}, rank={2:d}, nnz={3:d}, "
                   "time={4:.3e}")
                  .format(i, err, np.sum(s > 0), np.sum(S > 0), svd_time))
        if err < delta:
            break
        i += 1

    #if i >= maxiter:
        #logging.warn("convergence not reached in pcp")
    return L, S, (u, s, v)


def shrink(M, tau):
    sgn = np.sign(M)
    S = np.abs(M) - tau
    S[S < 0.0] = 0.0
    return sgn * S


def _svd(method, X, rank, tol, **args):
    rank = min(rank, np.min(X.shape))
    if method == "exact":
        return np.linalg.svd(X, full_matrices=False, **args)
    elif method == "sparse":
        if rank >= np.min(X.shape):
            return np.linalg.svd(X, full_matrices=False)
        u, s, v = svds(X, k=rank, tol=tol)
        u, s, v = u[:, ::-1], s[::-1], v[::-1, :]
        return u, s, v
    elif method == "randomized":
        u, s, v = randomized_svd(X, n_components=X.shape[0],
                                 n_iter=1,
                                 random_state=123) # HARDCODED SEED
        return u, s, v
    raise ValueError("invalid SVD method")



"""Contains basic signal processing utilities.

"""
from scipy.signal import butter, lfilter, freqz

def _butter_lowpass(cutoff, Fs, order=5):
    nyq = 0.5 * Fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff,
                    btype='low', analog=False)
    return b, a

def butter_lowpass_filter(data, cutoff, Fs, order=5):
    r"""Constructs a lowpass filter at given cutoff

    Parameters
    ----------
    data : numpy array 
        a `t` by 1 numpy array representing a discretely sampled time-series signal with `t` timesteps.

    cutoff : float
        the upper bound frequency let through the filter.
    Fs : float
        the sampling frequency of the signal
    order : int
        the order of the filter

    Returns
    -------
    filtered_signal : numpy matrix
        the filtered signal

    See Also
    --------
    butter_highpass_filter :
    butter_bandpass_filter :
    butter_bandstop_filter :

    Notes
    -----
    Relies on scipy.signal

    Examples
    --------

    >>> T = 10 # Signal lasts 10 seconds
    >>> Fs = 40 # We sample the signal at 40 Hz
    >>> w1 = .1
    >>> w2 = 1
    >>> w3 = 4
    >>> sample_points = np.linspace(0, T, T * Fs)
    >>> sines_w1 = np.sin(w1 * 2 * np.pi * sample_points)
    >>> sines_w2 = np.sin(w2 * 2 * np.pi * sample_points)
    >>> sines_w3 = np.sin(w3 * 2 * np.pi * sample_points)
    >>> signal = sines_w1 + sines_w2 + sines_w3
    >>> new_signal = butter_lowpass_filter(signal, 2, Fs, 5)

    """
    b, a = _butter_lowpass(cutoff, Fs, order=order)
    filtered_signal = lfilter(b, a, data)
    return filtered_signal


def _butter_highpass(cutoff, Fs, order=5):
    nyq = 0.5 * Fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff,
                btype='high', analog=False)
    return b, a

def butter_highpass_filter(data, cutoff, Fs, order=5):
    r"""Constructs a highpass filter at given cutoff

    Parameters
    ----------
    data : ``ndarray`` 
        a `(t, 1)` ``ndarray`` representing a discretely sampled time-series signal with `t` timesteps.

    cutoff : float
        the upper bound frequency let through the filter.
    Fs : float
        the sampling frequency of the signal
    order : int
        the order of the filter

    Returns
    -------
    filtered_signal : numpy matrix
        the filtered signal

    See Also
    --------
    butter_highpass_filter :
    butter_bandpass_filter :
    butter_bandstop_filter :

    Notes
    -----
    Relies on scipy.signal

    Examples
    --------

    >>> T = 10 # Signal lasts 10 seconds
    >>> Fs = 40 # We sample the signal at 40 Hz
    >>> w1 = .1
    >>> w2 = 1
    >>> w3 = 4
    >>> sample_points = np.linspace(0, T, T * Fs)
    >>> sines_w1 = np.sin(w1 * 2 * np.pi * sample_points)
    >>> sines_w2 = np.sin(w2 * 2 * np.pi * sample_points)
    >>> sines_w3 = np.sin(w3 * 2 * np.pi * sample_points)
    >>> signal = sines_w1 + sines_w2 + sines_w3
    >>> new_signal = butter_highpass_filter(signal, 2, Fs, 5)

    """
    b, a = _butter_highpass(cutoff, Fs, order=order)
    filtered_signal = lfilter(b, a, data)
    return filtered_signal

def _butter_bandpass(cutoff, Fs, order=5):
    nyq = 0.5 * Fs
    lowcut = cutoff[0] / nyq
    highcut = cutoff[1] / nyq
    b, a = butter(order, [lowcut, highcut],
                btype='bandpass', analog=False)
    return b, a

def butter_bandpass_filter(data, cutoff, Fs, order=5):
    r"""Constructs a bandpass filter at given cutoff range

    Parameters
    ----------
    data : ``ndarray`` 
        a `(t, 1)` ``ndarray`` representing a discretely sampled time-series signal with `t` timesteps.

    cutoff : [float, float]
        the [lower, upper] bound frequency let through the filter.
    Fs : float
        the sampling frequency of the signal
    order : int
        the order of the filter

    Returns
    -------
    filtered_signal : numpy matrix
        the filtered signal

    See Also
    --------
    butter_highpass_filter :
    butter_bandpass_filter :
    butter_bandstop_filter :

    Notes
    -----
    Relies on scipy.signal

    Examples
    --------

    >>> T = 10 # Signal lasts 10 seconds
    >>> Fs = 40 # We sample the signal at 40 Hz
    >>> w1 = .1
    >>> w2 = 1
    >>> w3 = 4
    >>> sample_points = np.linspace(0, T, T * Fs)
    >>> sines_w1 = np.sin(w1 * 2 * np.pi * sample_points)
    >>> sines_w2 = np.sin(w2 * 2 * np.pi * sample_points)
    >>> sines_w3 = np.sin(w3 * 2 * np.pi * sample_points)
    >>> signal = sines_w1 + sines_w2 + sines_w3
    >>> new_signal = butter_bandpass_filter(signal, [.6, 2], Fs, 5)

    """
    b, a = _butter_bandpass(cutoff, Fs, order=order)
    filtered_signal = lfilter(b, a, data)
    return filtered_signal

def _butter_bandstop(cutoff, Fs, order=5):
    nyq = 0.5 * Fs
    lowcut = cutoff[0] / nyq
    highcut = cutoff[1] / nyq
    b, a = butter(order, [lowcut, highcut],
                btype='bandstop', analog=False)
    return b, a

def butter_bandstop_filter(data, cutoff, Fs, order=5):
    r"""Constructs a bandstop filter at given cutoff

    Parameters
    ----------
    data : ``ndarray`` 
        a `(t, 1)` ``ndarray`` representing a discretely sampled time-series signal with `t` timesteps.

    cutoff : [float, float]
        the [lower, upper] bound frequency let stop with the filter.
    Fs : float
        the sampling frequency of the signal
    order : int
        the order of the filter

    Returns
    -------
    filtered_signal : numpy matrix
        the filtered signal

    See Also
    --------
    butter_highpass_filter :
    butter_bandpass_filter :
    butter_bandstop_filter :

    Notes
    -----
    Relies on scipy.signal

    Examples
    --------

    >>> T = 10 # Signal lasts 10 seconds
    >>> Fs = 40 # We sample the signal at 40 Hz
    >>> w1 = .1
    >>> w2 = 1
    >>> w3 = 4
    >>> sample_points = np.linspace(0, T, T * Fs)
    >>> sines_w1 = np.sin(w1 * 2 * np.pi * sample_points)
    >>> sines_w2 = np.sin(w2 * 2 * np.pi * sample_points)
    >>> sines_w3 = np.sin(w3 * 2 * np.pi * sample_points)
    >>> signal = sines_w1 + sines_w2 + sines_w3
    >>> new_signal = butter_bandstop_filter(signal, 2, Fs, 5)

    """
    b, a = _butter_bandstop(cutoff, Fs, order=order)
    filtered_signal = lfilter(b, a, data)
    return filtered_signal



import cPickle as pkl

# Import pipeline methods
import panda.methods.denoise as den
import panda.methods.bad_chans as bad_chans
import panda.methods.derivatives as der
import panda.methods.viz as viz
import panda.methods.interpolation as inter
import panda.methods.misc as misc

params = {
    'p_global': {
         'hpf': {
             'order': 6,
             'Fs': 500,
             'cutoff': 1
         },
         'lpf': {
             'order': 6,
             'Fs': 500,
             'cutoff': 230
         },
         'bsf': {
             'order': 6,
             'Fs': 500,
             'cutoffs': [[50, 70], 
                         [95, 105], 
                         [115, 125],
                         [175, 185],
                         [195, 205]]
         },
	'wave_visu': {
	    'wave': 'db2',
	    'verbose': False
	},
	'wave_rej': {
	    'wave': 'db2',
	    'thresh': 16,
            'rec': True
	},
        'amp_rej': {
            'thresh': 8,
            'method': 'shrink'
        },
        'rpca': {
            'max_iter': 60,
            'verbose': False,
            'pca_method': 'randomized',
            'delta': 1e-7,
            'mu': None
        },
        'bad_detec': {
            'thresh': 3,
            'measure': ['prob', 'kurtosis'],
            'normalize': 0,
            'discret': 1000,
            'verbose': False,
            'trim': None
        },
        'inter': {
            'loc_unit': 'radians',
                    'verbose': False,
                    'k': 3,
                    's' : 1000,
                    'wave': 'db2'
        },
        'sample_freq': 500,
        'eog_chans': [48, 49, 56, 63, 68,
                      73, 81, 88, 94, 99,
                      107, 113, 119, 125,
                      125, 126, 127, 128], # 0 indexed
        'plotting': {
            'time_compression': 10000,
            'notebook': False
        },
        'plot_folders': {
            'spectrogram_dir' : 'qa/spectrograms',
            'heatmap_dir' : 'qa/heatmaps',
            'correlation_dir' : 'qa/correlations',
            'sparkline_dir' : 'qa/sparklines'
        },
        'derivatives': {
            'correlation_matrix': 'derivatives/correlation_matrix',
            'spectrum': 'derivatives/spectrum',
            'coherence_matrix': 'derivatives/coherence_matrix',
            'left_singular_vectors': 'derivatives/left_singular_vectors'
        }
    },
    'functions': [misc.setup,
                  den.highpass,
#                  den.bandstop,
                  den.eog_regress,
                  den.wave_rejection,
                  den.amp_shrinkage,
                  bad_chans.bad_detec,
                  inter.wave_interp
                  ],

    'derivatives': [der.correlation],#, der.svd, der.coherence],
    'plots':[viz.heatmap, viz.spectrograms, viz.correlation]
}



from utils import execute_cmd
import sys
import boto3
import os

def get_data(bucket, remote_dir, local, subj=None, ses=None, public=True, folder = True):
    """
    Given an s3 bucket, data location on the bucket, and a download location,
    crawls the bucket and recursively pulls all data.
    """
    client = boto3.client('s3')
    if not public:
        bkts = [bk['Name'] for bk in client.list_buckets()['Buckets']]
        if bucket not in bkts:
            sys.exit("Error: could not locate bucket. Available buckets: " +
                     ", ".join(bkts))
    cmd = 'aws'
    if public:
        cmd += ' --no-sign-request'
    if folder:
        cmd = "".join([cmd, ' s3 cp --recursive s3://', bucket, '/',
                       remote_dir])
    else:
        cmd = "".join([cmd, ' s3 cp s3://', bucket, '/',
                       remote_dir])
    if subj is not None:
        cmd = "".join([cmd, '/sub-', subj])
        std, err = execute_cmd('mkdir -p ' + local + '/sub-' + subj)
        local += '/sub-' + subj

        if ses is not None:
            cmd = "".join([cmd, '/ses-', ses])
            std, err = execute_cmd('mkdir -p ' + local + '/ses-' + ses)
            local += '/ses-' + ses
    cmd = "".join([cmd, ' ', local])
    std, err = execute_cmd(cmd)
    return



import numpy as np

def frob(x, y):
    return np.linalg.norm(x-y)

# Diff Num 3 Cycles
def diff_num_3cycle(x, y):
    return np.abs(np.sum(x) - np.sum(y))

# Diff Num 4 Cycles
def diff_num_4cycle(x, y):
    def c4(A):
        return np.trace(np.linalg.matrix_power(A, 4)) + \
               np.trace(np.linalg.matrix_power(A, 2)) - \
               np.sum(np.sum(A, 0)**2)
    return np.abs(c4(x) - c4(y))

def diff_trace(x, y):
    return np.abs(np.trace(x) - np.trace(y))



import numpy as np

# Mean center and record max and min value in matrix (for plotting)
def setup(D, p_local, p_global):
    D = D - np.mean(D, axis = 1).reshape(-1, 1)
    zero_chans = []
    p_local['max'] = np.max(D)
    p_local['min'] = np.min(D)
    return (D, p_local)



import numpy as np
from utilities import apply_all
from copy import deepcopy
import networkx as nx

def loadGraphs(filenames, verb=False):
    """
    Given a list of files, returns a dictionary of graphs

    Required parameters:
        filenames:
            - List of filenames for graphs
    Optional parameters:
        verb:
            - Toggles verbose output statements
    """
    #  Initializes empty dictionary
    gstruct = OrderedDict()
    for idx, files in enumerate(filenames):
        if verb:
            print "Loading: " + files
        #  Adds graphs to dictionary with key being filename
        fname = os.path.basename(files)
        gstruct[fname] = nx.read_gpickle(files)
    return gstruct

def constructGraphDict(names, fs, verb=False):
    """
    Given a set of files and a directory to put things, loads graphs.

    Required parameters:
        names:
            - List of names of the datasets
        fs:
            - Dictionary of lists of files in each dataset
    Optional parameters:
        verb:
            - Toggles verbose output statements
    """
    #  Loads graphs into memory for all datasets
    graphs = OrderedDict()
    for idx, name in enumerate(names):
        if verb:
            print "Loading Dataset: " + name
        # The key for the dictionary of graphs is the dataset name
        graphs[name] = loadGraphs(fs[name], verb=verb)
    return graphs

def rdf(dist, ids):
    N = dist.shape[0]
    assert(N == len(ids))
    uniqids = list(set(ids))
    countvec = [ids.count(uniqid) for uniqid in uniqids]
    scans = np.max(countvec)
#     rdf = np.empty((N*(scans-1)))
    rdf = []

    for i in np.arange(0, N):
        ind = [idx for idx, x in enumerate(ids) if x == ids[i]]
        for j in ind:
            if i != j:
                di = deepcopy(dist[i,:])
                di[ind] = np.inf
                d = dist[i,j]
                diff = di[np.where(~np.isinf(di))]
#                 import pdb; pdb.set_trace()
                rdf += [1.0 - ((np.sum(diff < d) + 0.5*np.sum(diff == d)) / (1.0*(N-len(ind))))]
    return rdf

def partial_disc(D, labels, subject, trial1, trial2):
    enum = np.arange(D.shape[0])
    idx1 = [i for i, x in enumerate(labels) if x == subject]
    t1 = enum[idx1][trial1]
    t2 = enum[idx1][trial2]
    d_t1_t2 = D[t1][t2]
    
    idx2 = [i for i, x in enumerate(labels) if x != subject]
    d_ra = [D[t1][x] for x in enum[idx2]]
    
    return np.mean(d_t1_t2 < d_ra)

def distance_matrix(data, metric, symmetric = True):
    n = data.shape[2]
    dist_matrix = np.zeros((n, n))
    if symmetric:
        for i in range(n):
            for j in range(i):
                tmp = metric(data[:,:,i], data[:,:,j])
                dist_matrix[i][j] = tmp
                dist_matrix[j][i] = tmp
    else:
        for i in range(n):
            for j in range(n):
                dist_matrix[i][j] = metric(data[i], data[j])
    return dist_matrix

def discriminibility(data, labels, metric):
    dist_matrix = distance_matrix(data, metric)
    partials = []
    for s in list(set(labels)):
        num = ids.count(s)
        for t in range(num):
            for tt in range(num):
                if tt != t:
                    p = partial_disc(dist_matrix, labels, s, t, tt)
                    partials.append(p)
    return dist_matrix, np.mean(partials)

def disc_all(D, labels, transforms, metrics, names):
    discs = []
    derivitives = apply_all(D, transforms, map)
    for i in range(len(derivitives)):
        print names[i]
        disc = discriminibility(
                    derivitives[i],
                    labels,
                    metrics[i]
                )
        discs.append(disc)
        print disc
    return discs

def disc_plot(discs, disc_names, exp_names):
    import seaborn as sns
    import matplotlib.pyplot as plt
    discs = np.vstack(discs)
    print discs
    for disc_method in range(discs.shape[1]):
        plt.plot(discs[:, disc_method], '*', label = disc_names[disc_method])
    plt.xticks(range(len(exp_names)), exp_names)
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, 1.06),
          ncol=4, fancybox=True, shadow=True)
    plt.tight_layout()
    ax = plt.gca()
    ax.margins(0.05)
    return plt.gcf()

def intra_inter(dist, labels):
    within = []
    outside = []
    for i, label1 in enumerate(labels):
        for j, label2 in enumerate(labels):
            if label1 == label2:
                if i != j:
                    within.append(dist[i, j])
            else:
                outside.append(dist[i, j])
    return within, outside




from subprocess import Popen, PIPE
import sys


def execute_cmd(self, cmd):
    """
    Given a bash command, it is executed and the response piped back to the
    calling script
    """
    p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
    out, err = p.communicate()
    code = p.returncode
    if code:
        sys.exit("Error  " + str(code) + ": " + err)
    return out, err



import numpy as np
from pcp import pcp
np.seterr(divide='ignore')

# Return representation of the spectrum
def spect(D):
    return np.linalg.svd(D, full_matrices=False)[1]

# Return identity
def identity(D):
    return D

# Return correlation matrix
def correl(D):
    with np.errstate(invalid='ignore'):
        return np.nan_to_num(np.corrcoef(D))

# Return spectrum of the correlation matrix
def correl_spec(D):
    with np.errstate(invalid='ignore'):
        return np.linalg.svd(np.nan_to_num(np.corrcoef(D)), full_matrices=False)[1]

# Return thresholded correlation matrix
def thresh_correl(D, k):
    with np.errstate(invalid='ignore'):
        return np.nan_to_num(np.corrcoef(D)) > k - np.eye(D.shape[0])

def tc_make(k):
    def tc(D):
        return thresh_correl(D, k)
    return tc

def pca_singvals(d):
    m = np.mean(d, axis = 1).reshape(-1, 1)
    d_cent = d - m
    U, _, _ = np.linalg.svd(d_cent, full_matrices = False)
    return U

def rpca_singvals(d):
    return pca_singvals(pcp(d, maxiter=5, verbose=True, svd_method='randomized')[0])



from math import radians, cos, sin, asin, sqrt
import pywt
import numpy as np
import copy
from pcp import pcp as pcp
from signals import butter_bandstop_filter, butter_highpass_filter, butter_lowpass_filter

def eog_regress(D, p_local, p_global):
    p_local['eog_in'] = False
    Y, N = (D[p_global['eeg_chans'], :], D[p_global['eog_chans'], :])
    m_Y = np.mean(Y, axis = 1).reshape(-1, 1)
    Y = Y - m_Y
    m_N = np.mean(N, axis = 1).reshape(-1, 1)
    N = N - m_N
    return (Y - ((Y.dot(N.T)).dot(np.linalg.pinv(N.dot(N.T)))).dot(N), p_local)

def highpass(D, p_local, p_global):
    order = p_global['hpf']['order']
    Fs = p_global['hpf']['Fs']
    cutoff = p_global['hpf']['cutoff']
    one_chan = lambda i: butter_highpass_filter(D[i, :], cutoff, Fs, order)
    D = np.vstack(one_chan(i) for i in range(D.shape[0]))
    return (D, p_local)

def lowpass(D, p_local, p_global):
    order = p_global['lpf']['order']
    Fs = p_global['lpf']['Fs']
    cutoff = p_global['lpf']['cutoff']
    one_chan = lambda i: butter_lowpass_filter(D[i, :], cutoff, Fs, order)
    D = np.vstack(one_chan(i) for i in range(D.shape[0]))
    return (D, p_local)

def bandstop(D, p_local, p_global):
    order = p_global['bsf']['order']
    Fs = p_global['bsf']['Fs']
    for cutoff in p_global['bsf']['cutoffs']:
            one_chan = lambda i: butter_bandstop_filter(D[i, :],
							cutoff,
							Fs, order)
	    D = np.vstack(one_chan(i) for i in range(D.shape[0]))
    return (D, p_local)

def sure(t, X):
    x_gt = np.abs(X) > t
    y = np.ones(X.shape) * t
    y[x_gt] = np.abs(X)[x_gt]
    S = np.sum(np.square(y))
    n = X.shape[0]
    card = np.sum(np.logical_not(x_gt))
    return n + S -2 * card

def sure_shrink_fast(X):
    n = X.shape[0]
    g_thresh = np.sqrt(2 * np.log(n))
    X_via = X[np.abs(X) <= g_thresh]
    X_sort = np.sort(X_via)
    csX = np.cumsum(X_sort)
    R = np.arange(X_via.shape[0])
    L = map(lambda t: n - 2 * t + csX[t] + (n - t) * X_sort[t]**2, R)
    if len(L) == 0:
        min_thresh = g_thresh
    else:
        min_thresh = X_sort[np.argmin(L)]
    absX = np.abs(X)
    X[absX < min_thresh] = 0
    X[X > min_thresh] = X[X > min_thresh] - min_thresh
    X[X < -min_thresh] = X[X < -min_thresh] + min_thresh
    return (X, min_thresh)

def sure_shrink_denoise(f, wave, v):
    true_coefs = pywt.wavedec(f, wave, level=None)
    res = [sure_shrink_fast(coef) for coef in true_coefs[1:]]
    den_coefs, thresholds = zip(*res)
    den_coefs = list(den_coefs)
    den_coefs.insert(0, true_coefs[0])
    f_denoised = pywt.waverec(den_coefs, wave)
    return f_denoised

def wavelet_sureshrink(d, p_local, p_global):
    C = d.shape[0]
    T = d.shape[1]
    wave = p_global['wave_sure']['wave']
    v = p_global['wave_sure']['verbose']
    # Make sure even length timesteps
    if T % 2 == 1:
        d = d[:, :-1]

    # Get original coeffs for plot
    if v:
        coefs = [pywt.wavedec(d[c, :], wave) for c in range(C)]

    # Denoise each channel
    d_den = np.vstack(map(lambda i: sure_shrink_denoise(d[i, :], wave, v),
                  range(d.shape[0])))

    # Get new coeffs for plot
    if v:
        den_coefs = [pywt.wavedec(d_den[c, :], wave) for c in range(C)]
        from viz import cross_compare
        for i in range(3):#range(len(coefs[0]))[1:]:
            cross_compare(coefs, den_coefs, i)
    return (d_den, p_local)

def visu_shrink(X):
    n = X.shape[0]
    g_thresh = np.sqrt(2 * np.log(n))
    min_thresh = g_thresh
    absX = np.abs(X)
    X[absX < min_thresh] = 0
    X[X > min_thresh] = X[X > min_thresh] - min_thresh
    X[X < -min_thresh] = X[X < -min_thresh] + min_thresh
    return (X, min_thresh)

def visu_shrink_denoise(f, wave, v):
    true_coefs = pywt.wavedec(f, wave, level=2)
    res = [visu_shrink(coef) for coef in true_coefs]
    den_coefs, thresholds = zip(*res)
    den_coefs = list(den_coefs)
    #den_coefs.insert(0, true_coefs[0])
    f_denoised = pywt.waverec(den_coefs, wave)
    return f_denoised

def wavelet_visushrink(d, p_local, p_global):
    C = d.shape[0]
    T = d.shape[1]
    wave = p_global['wave_visu']['wave']
    v = p_global['wave_visu']['verbose']
    # Make sure even length timesteps
    if T % 2 == 1:
        d = d[:, :-1]

    # Get original coeffs for plot
    if v:
        coefs = [pywt.wavedec(d[c, :], wave) for c in range(C)]

    # Denoise each channel
    d_den = np.vstack(map(lambda i: visu_shrink_denoise(d[i, :], wave, v),
                  range(d.shape[0])))

    # Get new coeffs for plot
    if v:
        den_coefs = [pywt.wavedec(d_den[c, :], wave) for c in range(C)]
        from viz import cross_compare
        for i in range(3):#range(len(coefs[0]))[1:]:
            cross_compare(coefs, den_coefs, i)
    return (d_den, p_local)

def pca_denoise(D, p_local, p_global):
    d, s_vals = D
    k = p_global['pca_den']['k']
    U_k = s_vals[:, :k]
    P_k = U_k.dot(U_k.T)
    return P_k.dot(d - np.mean(d, axis=1).reshape(-1, 1))

def rpca_denoise(D, p_local, p_global):
    M = p_global['rpca']['max_iter']
    v = p_global['rpca']['verbose']
    method = p_global['rpca']['pca_method']
    delta = p_global['rpca']['delta']
    mu = p_global['rpca']['mu']
    return (pcp(D, mu = mu, delta = delta, maxiter=M, verbose=v, svd_method=method)[0], p_local)

def wave_rejection(D, p_local, p_global):
    sigs = []
    wave = p_global['wave_rej']['wave']
    thresh = p_global['wave_rej']['thresh']
    all_coefs = [pywt.wavedec(D[i, :], wave, level = None) for i in range(D.shape[0])]
    for i, coefs in enumerate(all_coefs):
        for j , coef in enumerate(coefs[1:]):
            mu = np.mean(coef)
            std = np.std(coef)
            coef = np.array(coef)
            coef[np.abs(coef) > mu + thresh * std] = 0
            all_coefs[i][j + 1] = coef
    for coef in all_coefs:    
        sigs.append(pywt.waverec(coef, 'db2'))
    if p_global['wave_rej']['rec']:
        return (np.vstack(sigs), p_local)
    else:
        return (all_coefs, p_local)

def amp_shrinkage(D, p_local, p_global):
    thresh = p_global['amp_rej']['thresh']
    method = p_global['amp_rej']['method']
    D = np.copy(D)
    std = np.std(D)
    mu = np.mean(D)
    high = D > mu + thresh * std
    low = D < mu - thresh * std
    if method == 'zero':
        D[high] = 0
        D[low] = 0
    elif method == 'shrink':
        D[high] = mu + thresh * std
        D[low] = mu - thresh * std
    return (D, p_local)



import sys
import csv
import numpy as np
import cPickle as pkl
import os
import traceback
from glob import glob as glob
import dataset_creation as dc
from pathos.multiprocessing import ProcessingPool as Pool
import matplotlib.pyplot as plt

def load_data(pkl_path):
    with open(pkl_path, 'rb') as f:
        return pkl.load(f)

def data_generator_factory(dataset):
    subjects, trials = zip(*dc.participant_info(dataset))
    trials = map(int, trials)
    subjects = np.array(subjects)
    zips = zip(subjects, trials)
    zips = [[x[0]] * x[1] for x in zips]
    labels = [item for sublist in zips for item in sublist]
    def data_gen():
        c = -1
        for i, subject in enumerate(subjects):
            try:
                sub_path = 'figs/' + subject 
                os.makedirs(sub_path)
            except:
                print 'Already exists figure folder for ', subject
            for trial in range(1, trials[i] + 1):
                trial = '%s_trial-%02d' % (subject, int(trial))
                fig_path = 'figs/' + subject + '/' + trial + '/'
                try:
                    os.makedirs(fig_path)
                except:
                    print 'Already exists figure folder for ', trial
                print 'OPENING FILE', trial
                pkl_path = '%s/%s/eeg/%s.pkl' % (dataset, subject, trial)
                p_local = {
                    'fig_path': fig_path,
                    'step': 0,
                    'function_name': 'Raw'
                }
                yield (load_data(pkl_path), p_local)

    def factory():
        return data_gen()
    return factory, labels

def f(x, function, params):
    print x
    x0 = x[0]()
    try:
        x[1]['step'] = x[1]['step'] + 1
        print x[1]
        return function(x0, x[1], params['p_global'])
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
    print "*** print_tb:"
    traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
    print "*** print_exception:"
    traceback.print_exception(exc_type, exc_value, exc_traceback,
              limit=2, file=sys.stdout)
    print "*** print_exc:"
    traceback.print_exc()

def ngf(factory, function, params):
    def nested_generator():
        for item in factory():
            yield item
    return nested_generator

def tgf(factory1, factory2):
    def tup_generator():
        for item in ((a, b) for (a, b) in zip(factory1(), factory2())):
            yield item
    return tup_generator

def ngff(init, functions, params):
    nfg = init
    for function in functions:
        nfg = ngf(nfg, function, params)
    return nfg

def capstone(D, p_local, p_global):
    def f():
        return D
    return (f, p_local)

def apply_all(D, functs, map_method):
    res = []
    for funct in functs[::-1]:
        res.append(map(funct, map_method(lambda d: d, D)))
    return res



import numpy as np
import cPickle as pkl
import scipy
import scipy.signal

def correlation(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    with np.errstate(invalid='ignore'):
        D = np.nan_to_num(np.corrcoef(D))

    path = p_global['derivatives']['correlation_matrix'] \
           + '/' + 'correlation_matrix'
    save_and_close(D, path, p_local)

def svd(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    U, s, _ = np.linalg.svd(D, full_matrices=False)

    path_spectrum = p_global['derivatives']['spectrum'] \
                    + '/' + 'spectrum'
    path_lsv = p_global['derivatives']['left_singular_vectors'] \
               + '/' + 'left_singular_vectors'
    save_and_close(s, path_spectrum, p_local)
    save_and_close(U, path_lsv, p_local)

def coherence(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    C = D.shape[0]
    coh = np.zeros([C, C])
    for i in range(C):
        for j in range(i):
            x = D[i, :]
            y = D[j, :]
            x_y_coh = scipy.signal.coherence(x, y, p_global['sample_freq']) 
            x_y_coh = np.mean(x_y_coh[1])
            coh[i, j] = x_y_coh
            coh[j, i] = x_y_coh
    coh = coh + np.eye(C)
    path = p_global['derivatives']['coherence_matrix'] \
               + '/' + 'coherence'
    save_and_close(coh, path, p_local)

def save_and_close(D, path, p_local):
    der = ("%s/%s_%s.pkl") % (p_local['out_path'], path, p_local['funct'])
    with open(der, 'wb') as f:
        pkl.dump(D, f, -1)



import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import seaborn as sns
import numpy as np
import scipy.signal
from scipy.signal import spectrogram

def spectrograms(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    C = D.shape[0]
    T = D.shape[1]
    for c in range(C)[:3]:
        f, t, Sxx = spectrogram(D[c, :], p_global['sample_freq'])
        sns.heatmap(np.log(Sxx[::-1, :]), xticklabels = t.astype(int),
            yticklabels = f.astype(int)[::-1])

        # There is probably a better way to do this
        for label in plt.gca().get_xticklabels():
            label.set_visible(False)
        for label in plt.gca().get_xticklabels()[::Sxx.shape[1] / 6]:
            label.set_visible(True)
        for label in plt.gca().get_yticklabels():
            label.set_visible(False)
        for label in plt.gca().get_yticklabels()[::Sxx.shape[0] / 6]:
            label.set_visible(True)
        cbar = plt.gca().collections[0].colorbar
        plt.title('Spectrogram for channel ' + str(c + 1))
        plt.xlabel('Time in seconds')
        plt.ylabel('Frequency')
        cbar.set_label(r"$\log(\hat{f})$", labelpad=20, rotation=270)
        path = p_global['plot_folders']['spectrogram_dir'] \
               + '/' + 'spectrogram-%03d' % (c + 1)
	if p_global['plotting']['notebook']:
	    show_and_close()
	else:
	    save_and_close(path, p_local)

def correlation(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    C = D.shape[0]

    with np.errstate(invalid='ignore'):
        D = np.nan_to_num(np.corrcoef(D))

    sns.heatmap(D, xticklabels = C / 10, yticklabels = C / 10, square = True)

    cbar = plt.gca().collections[0].colorbar
    cbar.set_label("Correlation Coefficient", labelpad=20, rotation=270)
    plt.title('Correlation Matrix')
    plt.xlabel('Channels')
    plt.ylabel('Channels')
    path = p_global['plot_folders']['correlation_dir'] \
           + '/' + 'correlation'
    if p_global['plotting']['notebook']:
        show_and_close()
    else:
        save_and_close(path, p_local)

def sym_log(c):
    def f(x):
        return np.sign(x) * np.log(1 + np.abs(x / c))
    return f

def heatmap(D, p_local, p_global):
    if p_local['eog_in']:
        D = D[p_global['eeg_chans'], :]
    C = D.shape[0]
    T = D.shape[1]
    plt.figure()
    compress = T / p_global['plotting']['time_compression']
    x_ticks = max(2, p_global['plotting']['time_compression'] / 10)
    y_ticks = max(2, C / 6)

    f, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(11, 8.5))
    sns.heatmap(data = D[:, ::compress],
                xticklabels = x_ticks,
                yticklabels = y_ticks,
                ax = ax1)
    plt.ylabel('channel')

    cbar = ax1.collections[0].colorbar
    cbar.set_label("raw", labelpad=20, rotation=270)

    sym_log_100 = sym_log(100)

    sns.heatmap(data = sym_log_100(D[:, ::compress]),
                xticklabels = x_ticks,
                yticklabels = y_ticks,
                ax = ax2)
    plt.ylabel('channel')
    cbar = ax2.collections[0].colorbar
    cbar.set_label("symlog100", labelpad=20, rotation=270)

    sym_log_5 = sym_log(5)

    sns.heatmap(data = sym_log_5(D[:, ::compress]),
                xticklabels = x_ticks,
                yticklabels = y_ticks,
                ax = ax3)

    cbar = ax3.collections[0].colorbar
    cbar.set_label("symlog5", labelpad=20,  rotation=270)
    path = p_global['plot_folders']['heatmap_dir'] \
           + '/' + 'heatmap'
    plt.ylabel('channel')
    plt.xlabel('time downsampled to 10,000 timesteps')
    if p_global['plotting']['notebook']:
        show_and_close()
    else:
        save_and_close(path, p_local)

def sparklines(D, p_local, p_global):
    sns.set_style('white')
    C = D.shape[0]
    T = D.shape[1]
    fig = plt.figure()
    fig.set_size_inches(8, .2 * C)
    compress = T / p_global['plotting']['time_compression']

    M = np.max(D)
    m = np.min(D)

    first = plt.subplot(C, 1, 1)
    plt.plot(D[0, ::compress], linewidth=0.75)
    ax = plt.gca()
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    plt.setp(first.get_xticklabels(), visible=False)
    plt.setp(first.get_yticklabels(), visible=False)
    plt.ylabel('1', fontsize=8, rotation = 0)
    for i in range(1, C):
        ax = plt.subplot(C, 1, i + 1, sharex = first)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)
        plt.plot(D[i, ::compress], linewidth=0.75)
        plt.ylim([m, M])
        plt.setp(ax.get_xticklabels(), visible=False)
        plt.setp(ax.get_yticklabels(), visible=False)
        plt.ylabel(str(i + 1), fontsize=8, rotation = 0)
    #plt.tight_layout()
    plt.xlim([0,p_global['plotting']['time_compression']])
    plt.subplots_adjust(wspace=0, hspace=0)
    path = p_global['plot_folders']['sparkline_dir'] \
           + '/' + 'sparklines'
    save_and_close(path, p_local)

def save_and_close(path, p_local):
    try:
        mpl.use('Agg')
    except:
        pass # already loaded
    img = ("%s/%s_%s.png") % (p_local['out_path'], path, p_local['funct'])
    plt.savefig(img)
    plt.cla()
    plt.clf()
    plt.close()


def show_and_close():
    plt.show()
    plt.cla()
    plt.clf()
    plt.close()



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm, kurtosis
import seaborn as sns
import sys
import traceback
import copy

def bad_detec(D, p_local, p_global):
    print "Test"
    D = np.copy(D)
    zero_chans = []
    for c in range(D.shape[0]):
        if np.sum(D[c, :]) == 0:
	    zero_chans.append(c)
    zero_chans = np.array(zero_chans, dtype=np.uint8)
    print zero_chans
    all_bcs = [zero_chans]
    for meas in p_global['bad_detec']['measure']:
        print "Here"
        chans = np.arange(D.shape[0])
        chans = np.setdiff1d(chans, p_global['eog_chans'])
        chans = np.setdiff1d(chans, zero_chans)
        statistics = get_statistic(
                           D,
                           meas,
                           p_global['bad_detec']['trim'],
                           p_global['bad_detec']['discret'],
                           p_global['bad_detec']['verbose'],
                           chans)
        thresh = p_global['bad_detec']['thresh']
        bad_chans = chans[np.logical_or(statistics < -thresh, statistics > thresh)]
        all_bcs.append(bad_chans)
        if p_global['bad_detec']['verbose']:
            bad_chan_plot(statistics, p_local, meas, thresh)
    if len(all_bcs) > 0:
        all_bcs = np.hstack(all_bcs)
        p_local.update({'bad_chans': all_bcs})
        D[all_bcs, :] = 0
    else:
        all_bcs = None
        p_local.update({'bad_chans': all_bcs})
    return (D, p_local)

def gaussian_normalize(samples):
    mu = np.mean(samples)
    var = np.var(samples)
    normalized = (samples - mu) / np.sqrt(var)
    return normalized

def gaussian_trim(samples, trim_factor):
    z_thresh = norm.ppf(trim_factor)
    samples = samples[samples > z_thresh]
    samples = samples[samples < -z_thresh]
    return samples
    
def qq_plot(normalized, title):
    M = np.max(normalized)
    m = np.min(normalized)
    width = M - m
    ticks = np.linspace(0, 1, normalized.shape[0])
    plt.plot(np.linspace(0, 1, normalized.shape[0]), np.sort(normalized))
    plt.plot(np.linspace(0, 1, normalized.shape[0]), ticks * width + m)
    plt.show()

def std_norm(x):
    return norm.ppf((1 / np.sqrt(2 * np.pi)) * np.exp(-(x)**2 / 2))

def get_cdf(array, ticks):
    return np.array(map(lambda t: np.mean(array < t), ticks))


def get_statistic(D, measure, trim, discret, verbose, chans):
    # get num chans and time
    channels = D.shape[0]
    timepts = D.shape[1]

    statistic = []

    for c in chans:
        ra = D[c, :]
        if measure == 'prob':
            ra = gaussian_normalize(ra)
            if trim is not trim:
                #qq_plot(ra, title = 'qq-plot before gaussian trim')
                ra = gaussian_trim(ra, .1)
                ra = gaussian_normalize(ra)
                #qq_plot(ra, title = 'qq-plot after gaussian trim')
            M = np.max(ra)
            m = np.min(ra)
            ticks = np.linspace(0, 1, discret) * (M - m) + m 
            true_cdf = np.array(map(lambda x: norm.cdf(x), ticks))
            sample_cdf = get_cdf(ra, ticks)
            #plt.plot(ticks, true_cdf, 'r.', label='true')
            #plt.plot(ticks, sample_cdf, 'b.', label='sample')
            #plt.legend()
            #plt.title('Sample CDF compared to normal')
            #plt.show()
            statistic.append(np.max(true_cdf - sample_cdf))
        elif measure == 'kurtosis':
            statistic.append(kurtosis(ra))
    muS = np.mean(statistic)
    sigS = np.sqrt(np.var(statistic))
    normalized_statistic = (statistic - muS) / sigS
    return normalized_statistic




from math import radians, cos, sin, asin, sqrt
import pywt
import numpy as np
import copy
from scipy.interpolate import SmoothSphereBivariateSpline

import math as m

def cart2sph(x,y,z):
    XsqPlusYsq = x**2 + y**2
    r = m.sqrt(XsqPlusYsq + z**2)               # r
    elev = m.atan2(z,m.sqrt(XsqPlusYsq)) + np.pi / 2     # theta
    az = m.atan2(y,x) + np.pi                           # phi
    return elev, az, r

def haversine(rad, lon1, lat1, lon2, lat2):
    # convert decimal degrees to radians 
    if rad == 'degrees':
        lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a)) 
    return c

def gc(coords, i, j, r):
    t, p, r = cart2sph(coords[i][0], coords[i][1], coords[i][2])
    t_, p_, r_ = cart2sph(coords[j][0], coords[j][1], coords[j][2])
    args = {'rad' : r, 'lon1' : t,
      'lat1' : p, 'lon2' : t_, 'lat2': p_}
    return haversine(**args)

def neighbors(chan, chan_locs, candidates, k, r):
    dist = np.array(map(lambda c: gc(chan_locs, chan, c, r), candidates))
    closest = candidates[np.argsort(dist)[:k]]
    dist = dist[np.argsort(dist)[:k]]
    tot = np.sum(dist)
    return dist, closest, tot

def wave_interp(D, p_local, p_global):
    bad_chans = p_local['bad_chans']
    eog_chans = p_global['eog_chans']
    if bad_chans.size == 0:
        return (D, p_local)
    C = D.shape[0]
    T = D.shape[1]
    chan_locs = p_global['inter']['chan_locs']
    wave = p_global['inter']['wave']
    r = p_global['inter']['loc_unit']
    v = p_global['inter']['verbose']
    k = p_global['inter']['k']

    # Transform to wavelet coefficients
    coefs = [pywt.wavedec(D[c, :], wave) for c in range(C)]
    candidates = np.setdiff1d(range(C), bad_chans)
    candidates = np.setdiff1d(range(C), eog_chans)
    for bc in bad_chans:
        dist, neigh, tot = neighbors(bc, chan_locs, candidates, k, r)
        num_levels = len(coefs[bc])
        if v:
            print 'bad chan =', bc, ',', 'closest =', neigh, ',',
            print 'distances =', dist, 'num levels =', num_levels
        for i in range(num_levels):
            coefs[bc][i] = np.sum([coefs[n][i] * dist[p]/tot for p, n in enumerate(neigh)], axis=0)
    D_int = np.vstack([pywt.waverec(c, wave) for c in coefs])
    return (D_int, p_local)

def ssinterp(D, p_local, p_global):
    bad_chans = p_local['bad_chans']
    if bad_chans.size == 0:
        return (D, p_local)
    coords = p_global['inter']['chan_locs']
    coords = map(lambda x: cart2sph(*x), zip(coords[:, 0], coords[:, 1], coords[:, 2]))
    coords = np.vstack(coords)
    s_val = p_global['inter']['s']
    D = intp(D, coords, bad_chans, p_global['eog_chans'],  s = s_val)
    return (D, p_local) 

def intp(D, coords, rm_idx, eog_chans, s=1000):
    old = D[rm_idx, :]
    samp_idx = np.setdiff1d(np.arange(D.shape[0]), rm_idx)
    samp_idx = np.setdiff1d(samp_idx, eog_chans)
    sample = D[samp_idx]
    F = []
    for i in range(sample.shape[1]):
        F_i = ssi(sample[:, i],
                  coords[samp_idx],
                  s = s)
        F.append(F_i)
    pred = []
    for idx in rm_idx:
        itp = []
        for i, f in enumerate(F):
            if f is None:
                itp.append(D[idx, i])
            else:
                itp.append(f(coords[idx, 0],
                          coords[idx, 1])[0][0])
        pred.append(np.array(itp))
    D[rm_idx] = np.array(pred)
    return D


def ssi(E, P, s):
    F = None
    try:
        F = SmoothSphereBivariateSpline(P[:, 0], P[:, 1],
	                    		E, s=s)
    except:
        F = None
    return F    





from distutils.core import setup
from setuptools import setup
import panda

VERSION = 1.7

setup(
    name='ndd-panda',
    packages=[
        'panda',
        'panda.methods',
        'panda.utils',
        'panda.bench'
    ],
    scripts = [
        'panda/scripts/panda_pipeline'
    ],
    version=VERSION,
    description='Neuro Data Design EEG Preprocessing Pipeline',
    author='Ryan Marren and Nitin Kumar',
    author_email='rmarren1@jhu.edu, nkumarcc@gmail.com',
    url='https://github.com/NeuroDataDesign/orange-panda',
    keywords=[
        'eeg',
        'pipeline'
    ],
    classifiers=[],
    install_requires=[  # We didnt put versions for numpy, scipy, b/c travis-ci
        'numpy',  # We use nump v1.10.4
        'scipy',  # We use 0.17.0
        'seaborn',
        'pywavelets',
        'boto3',
        'matplotlib',
    ]
)
